import {
  API,
  Characteristic,
  DynamicPlatformPlugin,
  Logger,
  PlatformAccessory,
  PlatformConfig,
  Service,
} from "homebridge";
import { ControlUnit, DeviceButton } from "./Accessories";
import { Sequence } from "./Accessories/Sequence";
import HarmonyDataProvider from "./DataProviders/HarmonyDataProvider";
import { IConfig, IControlUnit, IDeviceButton } from "./Models/Config";
import { ISequence } from "./Models/Config/ISequence";
import { HarmonyDevice } from "./Models/HarmonyDevice";
import { HarmonyHub } from "./Models/HarmonyHub";
import { PLATFORM_NAME, PLUGIN_NAME } from "./settings";

export class Platform implements DynamicPlatformPlugin {
  constructor(
    public readonly log: Logger,
    config: PlatformConfig,
    public readonly api: API
  ) {
    this.log.debug("Finished initializing platform:", config.name);

    this.config = config as unknown as IConfig;
    //construct data provider
    const dataProvider = new HarmonyDataProvider({
      hubs: this.config.Hubs,
      deviceConfigs: this.config.Devices,
      matrix: this.config.Matrix,
      log: this.log,
    });

    let didFinishLaunching = false;
    this.api.on("didFinishLaunching", async () => {
      log.debug("Executed didFinishLaunching callback");
      didFinishLaunching = true;
    });

    this.dataProvider = null;

    if (this.config) {
      //construct data provider
      this.dataProvider = new HarmonyDataProvider({
        hubs: this.config.Hubs,
        deviceConfigs: this.config.Devices,
        matrix: this.config.Matrix,
        log: this.log,
      });

      //Emit devices if requested

      this.dataProvider.on("Ready", () => {
        this.log.info("All hubs connected");
        this.discoverControlUnitAccessories(dataProvider);
        this.discoverDeviceButtonAccessories(dataProvider);
        this.discoverSequenceAccessories(dataProvider);
        this.pruneAccessories();

        if (this.config.EmitDevicesOnStartup) {
          const hubs = this.dataProvider!.hubs;
          Object.values(hubs).forEach((hub: HarmonyHub) => {
            const deviceDictionary = hub.devices;
            this.log.info(`${hub.hubName}`);

            Object.values(deviceDictionary).forEach((device: HarmonyDevice) => {
              this.log.info(`    ${device.name} : ${device.id}`);
              Object.keys(device.commands).forEach((command: string) => {
                this.log.info(`        ${command}`);
              });
            });
          });
        }
      });
    }
  }

  public readonly Service: typeof Service = this.api.hap.Service;
  public readonly Characteristic: typeof Characteristic =
    this.api.hap.Characteristic;

  // this is used to track restored cached accessories
  public readonly accessories: PlatformAccessory[] = [];
  public config: IConfig;
  public dataProvider: HarmonyDataProvider | null;

  /**
   * Discover new control unit accessories
   * @param dataProvider
   */
  private discoverControlUnitAccessories(
    dataProvider: HarmonyDataProvider
  ): void {
    this.config.ControlUnits.forEach((unit: IControlUnit) => {
      const uuid = this.api.hap.uuid.generate(unit.DisplayName);
      const existingAccessory = this.accessories.find((e) => e.UUID === uuid);

      if (existingAccessory) {
        this.log.info(
          "Restoring existing accessory from cache: " +
            existingAccessory.displayName
        );

        new ControlUnit(this, existingAccessory, dataProvider, unit.Activities);

        this.api.publishExternalAccessories(PLUGIN_NAME, [existingAccessory]);
        console.log("Publishing external accessory: " + uuid);
      } else {
        this.log.info("Adding new accessory: " + unit.DisplayName);
        const accessory = new this.api.platformAccessory(
          unit.DisplayName,
          uuid
        );
        accessory.context["DeviceName"] = unit.DisplayName;
        accessory.context["Type"] = typeof ControlUnit.name;

        new ControlUnit(this, accessory, dataProvider, unit.Activities);

        this.api.publishExternalAccessories(PLUGIN_NAME, [accessory]);
        console.log("Publishing external accessory: " + uuid);
      }
    });
  }

  /**
   * Discover new device button accessories
   * @param dataProvider
   */
  private discoverDeviceButtonAccessories(
    dataProvider: HarmonyDataProvider
  ): void {
    this.config.DeviceButtons.forEach((button: IDeviceButton) => {
      const uuid = this.api.hap.uuid.generate(button.DisplayName);
      const existingAccessory = this.accessories.find((e) => e.UUID === uuid);
      if (existingAccessory) {
        this.log.info(
          "Restoring existing accessory from cache: " +
            existingAccessory.displayName
        );

        new DeviceButton(this, existingAccessory, dataProvider, button);
        this.api.updatePlatformAccessories([existingAccessory]);
      } else {
        this.log.info("Adding new accessory: " + button.DisplayName);
        const accessory = new this.api.platformAccessory(
          button.DisplayName,
          uuid
        );
        accessory.context["DeviceName"] = button.DisplayName;
        accessory.context["Type"] = typeof DeviceButton;

        new DeviceButton(this, accessory, dataProvider, button);

        this.api.registerPlatformAccessories(PLUGIN_NAME, PLATFORM_NAME, [
          accessory,
        ]);
        this.accessories.push(accessory);
      }
    });
  }

  /**
   * Discover new sequence accessories
   * @param dataProvider
   */
  private discoverSequenceAccessories(dataProvider: HarmonyDataProvider): void {
    this.config.Sequences.forEach((sequence: ISequence) => {
      const uuid = this.api.hap.uuid.generate(sequence.DisplayName);
      const existingAccessory = this.accessories.find((e) => e.UUID === uuid);
      if (existingAccessory) {
        this.log.info(
          "Restoring existing accessory from cache: " +
            existingAccessory.displayName
        );

        new Sequence(this, existingAccessory, dataProvider, sequence);
        this.api.updatePlatformAccessories([existingAccessory]);
      } else {
        this.log.info("Adding new accessory: " + sequence.DisplayName);
        const accessory = new this.api.platformAccessory(
          sequence.DisplayName,
          uuid
        );
        accessory.context["DeviceName"] = sequence.DisplayName;
        accessory.context["Type"] = typeof "Sequence";

        new Sequence(this, accessory, dataProvider, sequence);

        this.api.registerPlatformAccessories(PLUGIN_NAME, PLATFORM_NAME, [
          accessory,
        ]);
        this.accessories.push(accessory);
      }
    });
  }

  private pruneAccessories(): void {
    // Remove devices not discovered
    const listOfConfiguredAccessories = [
      ...this.config.DeviceButtons.map((e) => e.DisplayName),
      ...this.config.Sequences.map((e) => e.DisplayName),
      ...this.config.ControlUnits.map((e) => e.DisplayName),
    ];
    for (const accessory of this.accessories) {
      if (
        listOfConfiguredAccessories.filter(
          (displayName) => displayName === accessory.displayName
        ).length === 0
      ) {
        this.api.unregisterPlatformAccessories(PLUGIN_NAME, PLATFORM_NAME, [
          accessory,
        ]);
      }
    }
  }

  configureAccessory(accessory: PlatformAccessory<Record<string, any>>): void {
    this.log.info("Loading accessory from cache:", accessory.displayName);

    // add the restored accessory to the accessories cache so we can track if it has already been registered
    this.accessories.push(accessory);
  }
}
